from const import FG_COLOURS, BG_COLOURS

DEFAULT_FG = "white"
DEFAULT_BG = "black"


class Field(object):
    """
    Fields encapsulate advanced features for table cells, like
    custom formatters, colour and the like.

    Colour, in particular, will affect a strings "length" if the terminal
    control sequences are included, as a standard str() method would do.

    To that end, we override the __len__() and __str__() methods.
    """
    def __init__(self, value, fg=None, bg=None, field_name=None, formatters=None):
        self.value = value
        self.name = field_name
        self._fg=fg.lower()
        self._bg=bg.lower()
        self.formatters = formatters

    def __len__(self):
        return len(self.value)

    def __str__(self):
        prefix = FG_COLOURS[self._fg] + BG_COLOURS[self._bg]
        suffix = "\033[0m"
        return prefix + str(self.value) + suffix

    @classmethod
    def from_column(cls, column, value):
        field_fg = DEFAULT_FG
        field_bg = DEFAULT_BG
        if "fg" in column:
            field_fg = column["fg"]
        if "bg" in column:
            field_bg = column["bg"]
        return Field(value, fg=field_fg, bg=field_bg)


"""
Unit tests for texttab basic functionality
"""
import pytest

from texttab.table import BasicTable


# test_header_minimum_width
@pytest.mark.parametrize(
    "test_columns, expected_width",
    [
        ##################
        ([
            {"label": "Column 1", },
            {"label": "Value", },
        ], 20),
        ##################
        ([
            {
                "label": "Column 2",
                "width": 14,
            },
            {
                "label": "Value",
            }
        ], 24),
        ##################
        ([
            {
                "label": "Column 3",
                "width": 14,
            },
            {
                "label": "Value",
                "width": 10,
            }
        ], 27),
    ])
def test_header_minimum_width(test_columns, expected_width):
    """
    Minimum width (ie: no second field given in tuple) will be:
      Tw + 2N + (N-1) + 2
      Tw = Text width
      2N = Number of columns * 2 (for text spacing)
      +2 = constant for left and right edges
    """
    bt = BasicTable(columns=test_columns)
    assert bt.width == expected_width


@pytest.mark.parametrize(
    "test_columns, expected", [
        (
            [  # Testing centre alignment
                {
                    "label": "Column 3",
                    "width": 14,
                    "align": "center",
                },
                {
                    "label": "Value",
                    "width": 10,
                    "align": "center"
                }
            ],
            u"\u2502   Column 3   \u2502  Value   \u2502"
        ),
        ##################
        (
            [  # Testing centre alignment, mixed spellings, both valid
                {
                    "label": "Column 3",
                    "width": 14,
                    "align": "centre",
                },
                {
                    "label": "Value",
                    "width": 10,
                    "align": "center"
                }
            ],
            u"\u2502   Column 3   \u2502  Value   \u2502"
        ),
        ##################
        (
            [  # Testing right alignment
                {
                    "label": "Column 3",
                    "width": 14,
                    "align": "right",
                },
                {
                    "label": "Value",
                    "width": 10,
                    "align": "right"
                }
            ],
            u"\u2502     Column 3 \u2502    Value \u2502"
        ),
        ##################
        (
            [  # Testing right alignment, mixed case for alignment values
                {
                    "label": "Column 3",
                    "width": 14,
                    "align": "RIGHT",
                },
                {
                    "label": "Value",
                    "width": 10,
                    "align": "rIGHt"
                }
            ],
            u"\u2502     Column 3 \u2502    Value \u2502"
        )])
def test_header_label_alignments(test_columns, expected):
    bt = BasicTable(columns=test_columns)
    assert bt.generate_header_line() == expected


def test_minimum_width_autocorrects():
    """
    Making sure that if a width is specified that less than the text length
    of the label given, that the minimum width is calculated appropriately
    """
    test_columns = [
        {
            "label": "Column 3",
            "width": 4,
        },
        {
            "label": "Value",
            "width": 4,
        }
    ]
    bt = BasicTable(columns=test_columns)

    assert bt.width == 20


def test_data_row_min_fields_required():
    test_columns = [
        {"label": "Col 1"},
        {"label": "Column 2"},
        {"label": "Value"},
    ]

    test_data = ("foo", "bar",)
    bt = BasicTable(columns=test_columns)

    with pytest.raises(ValueError):
        bt.add_row(test_data)


def test_data_row_suitable_type_given():
    test_columns = [
        {"label": "Col 1"},
        {"label": "Column 2"},
        {"label": "Value"},
    ]

    test_data = "I am a string, I should be a list or tuple"
    bt = BasicTable(columns=test_columns)

    with pytest.raises(TypeError):
        bt.add_row(test_data)



"""
Formatter classes that can be applied to columns
"""
import stat

class ColumnFormatter(object):
    def __init__(self):
        return

    def format(self, value):
        raise NotImplemented


class ReadableBytesFormatter(object):
    @classmethod
    def format(cls, value, column):
        format_value = cls.human_readable_bytes(value)
        return "", format_value, False

    def human_readable_bytes(byteval):
        if byteval / 2**40 >= 1:
            return "{:>2.1f} TB ".format(byteval / 2**40)
        elif byteval / 2**30 >= 1:
            return "{:>2.1f} GB ".format(byteval / 2**30)
        elif byteval / 2**20 >= 1:
            return "{:>2.1f} MB ".format(byteval / 2**20)
        elif byteval / 2**10 >= 1:
            return "{:>2.1f} KB ".format(byteval / 2**10)
        else:
            return "{:>3.0f} B ".format(byteval)


class UnixPermissionsFormatter(ColumnFormatter):
    @classmethod
    def get_mode_char(cls, value):
        if stat.S_ISLNK(value):
            return 'l'
        elif stat.S_ISDIR(value):
            return 'd'
        elif stat.S_ISCHR(value):
            return 'c'
        elif stat.S_ISBLK(value):
            return 'b'
        elif stat.S_ISSOCK(value):
            return 's'
        elif stat.S_ISFIFO(value):
            return '|'
        else:
            return '-'

    @classmethod
    def format(cls, value, column):
        d = {
            0: '---',
            1: '--x',
            2: '-w-',
            3: '-wx',
            4: 'r--',
            5: 'r-x',
            6: 'rw-',
            7: 'rwx',
        }
        mode_char = cls.get_mode_char(value)
        ret_str = "{}{}{}{}".format(
            mode_char,
            d[(value & 0o0700) >> 6],
            d[(value & 0o0070) >> 3],
            d[value & 0o0007]
        )
        return "", ret_str, False



"""
Basic text-based table classes
"""
import const
from const import FG_COLOURS, BG_COLOURS
from fields import Field


class BasicTable(object):
    def __init__(self, columns=None, border="single", fg=None, bg=None,
                 head_fg=None, head_bg=None):
        self.__col_width_calculated = False

        self.columns = columns
        self.num_columns = len(self.columns)
        self.width = self.calculate_table_width()
        self.generate_column_labels()
        self.rows = []
        self.fg = fg
        self.bg = bg
        self.head_fg = head_fg
        self.head_bg = head_bg

        if border not in const.AVAILABLE_BORDERS:
            exception_string = "Unknown border style: '" + border + "'"
            raise ValueError(exception_string)
        self.border_style = border
        self.border_symbols = const.border_symbols[self.border_style]

    def render(self):
        table_lines = []
        table_lines.append(self._gen_header_top())
        table_lines.append(self.generate_header_line())
        table_lines.append(self._gen_header_bottom())

        for row in self.rows:
            table_lines.append(self._gen_table_row(row))

        table_lines.append(self._gen_table_bottom())

        return table_lines

    def add_row(self, rowdata):
        """
        Provide data to fill in the columns.
        rowdata must be a list or a tuple so that ordering is guaranteed.
        """
        if not isinstance(rowdata, (type(tuple([])), type([]))):
            raise TypeError(
                "Row data can only be expressed as a tuple or list")

        if len(rowdata) < self.num_columns:
            raise ValueError(
                "Insufficient number of data items. "
                "Received {}, expected {}".format(
                    len(rowdata), self.num_columns))
        elif len(rowdata) > self.num_columns:
            raise ValueError(
                "Too many data items. "
                "Received {}, expected {}".format(
                    len(rowdata), self.num_columns))
        self.rows.append(rowdata)

    def calculate_column_widths(self):
        for col in self.columns:
            minimum_width = len(col['label'].strip()) + 2
            if "width" not in col.keys():  # Auto calculate min width
                width = minimum_width
                col['width'] = width
            else:
                if col['width'] < minimum_width:
                    col['width'] = minimum_width

        self.__col_width_calculated = True

    def generate_column_labels(self):
        for col in self.columns:
            generated_label = ""
            if "align" in col.keys():
                if col['align'].lower() in ("center", "centre"):
                    fmt_str = " {:^" + str(col['width'] - 2) + "s} "
                elif col['align'].lower() == "left":
                    fmt_str = " {:" + str(col['width'] - 2) + "s} "
                elif col['align'].lower() == "right":
                    fmt_str = " {:>" + str(col['width'] - 2) + "s} "
            else:
                fmt_str = " {:" + str(col['width'] - 2) + "s} "

            if "head_fg" in col.keys():
                generated_label += const.FG_COLOURS[col["head_fg"]]
            elif "fg" in col.keys():
                generated_label += const.FG_COLOURS[col["fg"]]

            if "head_bg" in col.keys():
                generated_label += const.BG_COLOURS[col["head_bg"]]
            elif "bg" in col.keys():
                generated_label += const.BG_COLOURS[col["bg"]]

            generated_label += fmt_str.format(col['label'])
            generated_label += "\033[0m"
            col['gen_label'] = generated_label

    def calculate_table_width(self):
        if not self.__col_width_calculated:
            self.calculate_column_widths()

        table_width = 0
        for col in self.columns:
            table_width += col['width']

        table_width += 2 + (self.num_columns - 1)
        return table_width

    def get_colour_prefix(self):
        colour_string = ""
        reset_colour = False

        if self.fg is not None:
            colour_string += FG_COLOURS[self.fg]
            reset_colour = True
        if self.bg is not None:
            colour_string += BG_COLOURS[self.bg]
            reset_colour = True

        return colour_string, reset_colour

    def get_head_colour_prefix(self):
        colour_string = ""
        reset_colour = False

        if self.head_fg is not None:
            colour_string += FG_COLOURS[self.head_fg]
            reset_colour = True
        if self.head_bg is not None:
            colour_string += BG_COLOURS[self.head_bg]
            reset_colour = True

        return colour_string, reset_colour


    def generate_header_line(self):
        line = ""
        colour_string, reset_colours = self.get_head_colour_prefix()

        line += colour_string + self.border_symbols['VBAR']
        column_labels = [col['gen_label'] for col in self.columns]
        line += (colour_string + self.border_symbols["VBAR"]).join(column_labels)
        line += colour_string + self.border_symbols["VBAR"]

        if reset_colours is True:
            line += "\033[0m"
        return line

    def _gen_header_top(self):
        line = ""
        reset_colours = False

        if self.head_fg is not None:
            line += FG_COLOURS[self.head_fg]
            reset_colours = True
        if self.head_bg is not None:
            line += BG_COLOURS[self.head_bg]
            reset_colours = True

        line += self.border_symbols["TOP_LEFT"]
        col_bars = [
            col['width'] * self.border_symbols["HBAR"] for col in self.columns
        ]
        line += self.border_symbols["TOP_TEE"].join(col_bars)
        line += self.border_symbols["TOP_RIGHT"]

        if reset_colours is True:
            line += "\033[0m"
        return line

    def _gen_header_bottom(self):
        line = ""
        colour_string, reset_colours = self.get_head_colour_prefix()

        line += colour_string  # Begin our colour, column value might overwrite
        line += self.border_symbols["LEFT_TEE"]
        col_bars = [
            col['width'] * self.border_symbols["HBAR"] for col in self.columns
        ]
        line += colour_string + self.border_symbols["INTERSECT"].join(col_bars)
        line += colour_string + self.border_symbols["RIGHT_TEE"]

        if reset_colours is True:
            line += "\033[0m"

        return line

    def _gen_table_bottom(self):
        colour_string, reset_colours = self.get_colour_prefix()

        line = colour_string + self.border_symbols["BOTTOM_LEFT"]
        col_bars = [
            col['width'] * self.border_symbols["HBAR"] for col in self.columns
        ]
        line += (colour_string + self.border_symbols["BOTTOM_TEE"]).join(col_bars)
        line += colour_string + self.border_symbols["BOTTOM_RIGHT"]

        if reset_colours is True:
            line = line + "\033[0m"
        return line

    def _gen_table_row(self, rowdata):
        colour_string, reset_colours = self.get_colour_prefix()

        line = colour_string + self.border_symbols["VBAR"]
        col_strings = []

        for col_index, col in enumerate(self.columns):
            fmt_str = self._get_column_format_string(col)
            col_string = self.format_column_value(
                col, rowdata[col_index], fmt_str)
            col_strings.append(col_string)
        line += (colour_string + self.border_symbols["VBAR"]).join(col_strings)
        line += colour_string + self.border_symbols["VBAR"]

        if reset_colours is True:
            line += "\033[0m"

        return line

    def format_column_value(self, col, rowdata, fmtstr):
        # cell_field = Field.from_column(col, rowdata)
        ret_str = ""
        colour_string = ""
        reset_colour = False

        # Prefix our colour options, as ANSI codes
        if "fg" in col.keys():
            colour_string += FG_COLOURS[col["fg"]]
            reset_colour = True
        if "bg" in col.keys():
            colour_string += BG_COLOURS[col["bg"]]
            reset_colour = True
        if "bold" in col.keys():
            colour_string += "\033[1m"
            reset_colour = True
        if "invert" in col.keys():
            colour_string += "\033[7m"
            reset_colour = True

        # Get our width-padded value string, no colour modifiers to affect
        # length calculations.
        if 'formatter' in col.keys():
            colour_string, value_str, reset_colour = col['formatter'].format(rowdata, col)
        else:
            value_str = str(rowdata)
            # value_str = str(cell_field)
            if len(value_str) > col['width'] - 2:
                value_str = value_str[:col['width'] - 5] + "..."
        ret_str += colour_string + fmtstr.format(value_str)

        # Reset ANSI sequence if we tweaked any colour options
        if reset_colour is True:
            ret_str += "\033[0m"

        return ret_str

    def _get_column_format_string(self, col):
        if "align" not in col.keys() or col["align"] == "left":
            fmt_str = "{:" + str(col['width']) + "}"
        elif col["align"] == "right":
            fmt_str = "{:>" + str(col['width']) + "}"
        elif col["align"] in ("center", "centre"):
            fmt_str = "{:^" + str(col['width']) + "}"
        return fmt_str



from distutils.core import setup
setup(
    name='texttab',
    packages=['texttab'],
    version='0.5.1',
    license='MIT',
    description='Create ASCII tables that are flexible. Colours, various border styles and custom column formatters are some of the features.',   # Give a short description about your library
    author='Owen Klan',
    author_email='owen.j.klan@gmail.com',
    url='https://github.com/owenjklan/texttab',
    download_url='https://github.com/owenjklan/texttab/archive/refs/tags/v0.5.1.tar.gz',
    keywords=['command-line', 'ascii', 'tables', 'formatters', 'extensible'],
    install_requires=[
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        # Because we heavily use f-strings, a python 3.6 feature
        'Programming Language :: Python :: 3.6',
    ],
)